/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 950 $ $LastChangedDate: 2002-09-12 23:10:49 +0200 (Thu, 12 Sep 2002) $
 *
 * Runtime environment for gont.
 */

#include <gont/runtime.h>
#include <errno.h>
#include <iconv.h>
#include <stdlib.h>
#include <string.h>

struct gont_charset {
	iconv_t to;
	iconv_t from;
	char *name;
	gont_int approx;
	struct gont_lambda_expr *approx_fnc;
};

struct gont_charset *gont_default_charset, *gont_default_charset_with_approx;

struct gont_string *gont_string_utf8(const char *str)
{
	return gont_string_of_utf8(str, strlen(str));
}

/* Blah... we can't use extern on gont_string_of_utf8, since compiler
   prototypes it before.  */
struct gont_string *gont_string_of_utf8_(const char *str, gont_int len)
{
	return gont_string_of_utf8(str, len);
}

struct gont_string *gont_string_locale(const char *str)
{
	/* If we lack default charset fallback to iso-8859-1.  The main point
	   behind it is that it is equal to lower 256 chars of unicode, and
	   therfore we won't lose anything converting to/from it.  */
	if (gont_default_charset == NULL)
		gont_default_charset = gont_make_charset("iso-8859-1", 0, 0);
		
	if (gont_default_charset)
		return gont_string_of_cstring(str, strlen(str), 
					      gont_default_charset);
	else
		/* Oops, failed to open iconv converter for iso-8859-1, 
		   fallback to utf8.  */
		return gont_string_of_utf8(str, strlen(str));
}

void gont_free_charset(struct gont_charset *ch)
{
	if (ch->to != (iconv_t)(-1)) {
		iconv_close(ch->to);
		iconv_close(ch->from);
		ch->to = (iconv_t)(-1);
		ch->from = (iconv_t)(-1);
	}
}

static void finalize_charset(void *ch, void *dummy)
{
	gont_free_charset(ch);
	(void)dummy;
}

struct gont_charset *gont_make_charset(const char *ch, gont_int do_approx,
				       struct gont_lambda_expr *approx_fnc)
{
	struct gont_charset *ret;

	ret = gont_xmalloc(sizeof(*ret));
	ret->to = iconv_open(ch, "utf-8");
	if (ret->to == (iconv_t)(-1))
		return NULL;
	ret->from = iconv_open("utf-8", ch);
	if (ret->from == (iconv_t)(-1)) {
		iconv_close(ret->to);
		return NULL;
	}

	gont_register_finalizer(ret, finalize_charset, NULL);

	ret->name = gont_xmalloc_atomic(strlen(ch) + 1);
	strcpy(ret->name, ch);

	ret->approx = do_approx;
	ret->approx_fnc = approx_fnc;

	return ret;
}

void gont_charset_set_approx(struct gont_charset *ch, gont_int do_approx,
			     struct gont_lambda_expr *approx_fnc)
{
	ch->approx = do_approx;
	ch->approx_fnc = approx_fnc;
}

#define BUF_SIZE 4096

static unsigned char *
perform_iconv(iconv_t icnv, unsigned char *buf_hint, gont_int hint_size,
	      const unsigned char *src, gont_int src_size, gont_int *sizep,
	      gont_int do_approx, struct gont_lambda_expr *approx_fnc)
{
	char *in_ptr, *out_ptr;
	size_t in_left, out_left, ret;
	unsigned char *allocated_buf;
	gont_int buf_size;
	char scratch[BUF_SIZE];
	
	/* Reset iconv.  */
	iconv(icnv, NULL, NULL, NULL, NULL);
	
	in_ptr = (char*)src;
	in_left = src_size;

	buf_size = 0;

	if (hint_size) {
		hint_size--;
		out_left = hint_size;
		out_ptr = buf_hint;
	}

	for (;;) {
		size_t save_out;
		
		if (hint_size == 0) {
			out_left = BUF_SIZE;
			out_ptr = scratch;
		}
		
		save_out = out_left;
		ret = iconv(icnv, &in_ptr, &in_left, &out_ptr, &out_left);
		buf_size += save_out - out_left;
		
		if (ret == (size_t)(-1)) {
			switch (errno) {
			case EILSEQ:
				if (do_approx) {
					save_out = out_left;
					gont_append_approx(approx_fnc,
						     	   &in_ptr, &in_left, 
						     	   &out_ptr, &out_left);
					buf_size += save_out - out_left;
					if (errno == 0)
						continue;
					else
						break;
				}
				return NULL;
			case EINVAL:
				return NULL;
			case E2BIG:
				/* Need alloc.  */
				break;
			default:
				gont_sys_error(errno);
			}
		} else {
			if (hint_size) {
				if (sizep)
					*sizep = buf_size;
				if (in_left != 0)
					abort();
				buf_hint[buf_size] = '\0';
				return buf_hint;
			} else {
				break;
			}
		}

		/* Sorry, using hint not possible.  */
		hint_size = 0;
	}

	allocated_buf = gont_xmalloc_atomic(buf_size + 1);
	
	iconv(icnv, NULL, NULL, NULL, NULL);
	
	in_ptr = (char*)src;
	in_left = src_size;
	
	out_left = buf_size;
	out_ptr = allocated_buf;

	for (;;) {
		ret = iconv(icnv, &in_ptr, &in_left, &out_ptr, &out_left);

		if (ret != (size_t)(-1))
			break;
		
		switch (errno) {
		case EILSEQ:
			if (do_approx) {
				gont_append_approx(approx_fnc,
						   &in_ptr, &in_left, 
						   &out_ptr, &out_left);
				if (errno)
					abort();
				break;
			}
		case EINVAL:
		case E2BIG:
			abort();	/* WTF? */
		default:
			/* Out of memory or sth like that is possible.  */
			gont_sys_error(errno);
		}
	}
	
	if (out_left != 0 || in_left != 0)
		abort();

	if (sizep)
		*sizep = buf_size;
	
	allocated_buf[buf_size] = '\0';

	return allocated_buf;
}

unsigned char *gont_string_to_cstring(unsigned char *buf_hint, 
				      gont_int hint_size,
				      struct gont_string *s,
				      gont_int *sizep,
				      struct gont_charset *ch)
{
	unsigned char *utf;
	gont_int utf_size;
	unsigned char buf[BUF_SIZE];
	unsigned char *ret;
	
	if (ch->to == (iconv_t)(-1))
		gont_invalid_argument("gont_string_to_cstring");

	utf = gont_string_to_utf8(buf, sizeof(buf), s, &utf_size);

	ret = perform_iconv(ch->to, buf_hint, hint_size, 
			    utf, utf_size, sizep, ch->approx, ch->approx_fnc);

	if (ret == NULL) {
		if (errno == EILSEQ) {
			gont_character_not_supported_by_encoding();
		} else {
			abort();
		}
	}
	
	return ret;
}

struct gont_string *gont_string_of_cstring(const unsigned char *str, gont_int sz,
					   struct gont_charset *ch)
{
	gont_int ret_size;
	unsigned char buf[BUF_SIZE], *ret;
	
	if (ch->from == (iconv_t)(-1))
		gont_invalid_argument("gont_string_of_cstring");
	
	ret = perform_iconv(ch->from, buf, BUF_SIZE, str, sz, &ret_size, 0, 0);

	if (ret == NULL)
		switch (errno) {
		case EILSEQ:
			gont_invalid_mbseq();
			break;
		case EINVAL:
			gont_incomplete_mbseq();
			break;
		default:
			abort();
		}

	return gont_string_of_utf8(ret, ret_size);
}
